@@ -149,6 +149,11 @@ int exit_status;
 boolean ignore_readdir_race;
 
 
+/* If true, we issue warning messages
+ */
+boolean warnings;
+
+
 #ifdef DEBUG_STAT
 static int
 debug_stat (file, bufp)
@@ -176,6 +181,16 @@ main (int argc, char **argv)
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
+  if (isatty(0))
+    {
+      warnings = true;
+    }
+  else
+    {
+      warnings = false;
+    }
+  
+  
   predicates = NULL;
   last_pred = NULL;
   do_dir_first = true;
@@ -317,7 +332,17 @@ main (int argc, char **argv)
   for (i = 1; i < argc && strchr ("-!(),", argv[i][0]) == NULL; i++)
     process_top_path (argv[i]);
   if (i == 1)
-    process_top_path (".");
+    {
+      /* 
+       * We use a temporary variable here because some actions modify 
+       * the path temporarily.  Hence if we use a string constant, 
+       * we get a coredump.  The best example of this is if we say 
+       * "find -printf %H" (note, not "find . -printf %H").
+       */
+      char defaultpath[2] = ".";
+      process_top_path (defaultpath);
+    }
+  
 
   return exit_status;
 }
