@@ -17,7 +17,6 @@
    USA.
 */
 
-#define _GNU_SOURCE
 #include "defs.h"
 
 #include <fnmatch.h>
@@ -160,6 +159,7 @@ struct pred_assoc pred_table[] =
   {pred_cnewer, "cnewer  "},
   {pred_comma, ",       "},
   {pred_ctime, "ctime   "},
+  {pred_delete, "delete  "},
   {pred_empty, "empty   "},
   {pred_exec, "exec    "},
   {pred_false, "false   "},
@@ -245,26 +245,42 @@ struct prec_assoc prec_table[] =
  
    Return true if the file passes this predicate, false if not. */
 
-boolean
-pred_amin (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+
+/* pred_timewindow
+ *
+ * Returns true if THE_TIME is 
+ * COMP_GT: after the specified time
+ * COMP_LT: before the specified time
+ * COMP_EQ: less than WINDOW seconds after the specified time.
+ */
+static boolean
+pred_timewindow(time_t the_time, struct predicate const *pred_ptr, int window)
 {
   switch (pred_ptr->args.info.kind)
     {
     case COMP_GT:
-      if (stat_buf->st_atime > (time_t) pred_ptr->args.info.l_val)
-	return (true);
+      if (the_time > (time_t) pred_ptr->args.info.l_val)
+	return true;
       break;
     case COMP_LT:
-      if (stat_buf->st_atime < (time_t) pred_ptr->args.info.l_val)
-	return (true);
+      if (the_time < (time_t) pred_ptr->args.info.l_val)
+	return true;
       break;
     case COMP_EQ:
-      if ((stat_buf->st_atime >= (time_t) pred_ptr->args.info.l_val)
-	  && (stat_buf->st_atime < (time_t) pred_ptr->args.info.l_val + 60))
-	return (true);
+      if ((the_time >= (time_t) pred_ptr->args.info.l_val)
+	  && (the_time < (time_t) pred_ptr->args.info.l_val + window))
+	return true;
       break;
     }
-  return (false);
+  return false;
+}
+
+
+boolean
+pred_amin (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+{
+  (void) &pathname;
+  return pred_timewindow(stat_buf->st_atime, pred_ptr, 60);
 }
 
 boolean
@@ -295,6 +311,8 @@ pred_and (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 boolean
 pred_anewer (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
+  (void) &pathname;
+  
   if (stat_buf->st_atime > pred_ptr->args.time)
     return (true);
   return (false);
@@ -303,57 +321,32 @@ pred_anewer (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 boolean
 pred_atime (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
-  switch (pred_ptr->args.info.kind)
-    {
-    case COMP_GT:
-      if (stat_buf->st_atime > (time_t) pred_ptr->args.info.l_val)
-	return (true);
-      break;
-    case COMP_LT:
-      if (stat_buf->st_atime < (time_t) pred_ptr->args.info.l_val)
-	return (true);
-      break;
-    case COMP_EQ:
-      if ((stat_buf->st_atime >= (time_t) pred_ptr->args.info.l_val)
-	  && (stat_buf->st_atime < (time_t) pred_ptr->args.info.l_val
-	      + DAYSECS))
-	return (true);
-      break;
-    }
-  return (false);
+  (void) &pathname;
+  return pred_timewindow(stat_buf->st_atime, pred_ptr, DAYSECS);
 }
 
 boolean
 pred_close (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
+  (void) &pathname;
+  (void) &stat_buf;
+  (void) &pred_ptr;
+  
   return (true);
 }
 
 boolean
 pred_cmin (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
-  switch (pred_ptr->args.info.kind)
-    {
-    case COMP_GT:
-      if (stat_buf->st_ctime > (time_t) pred_ptr->args.info.l_val)
-	return (true);
-      break;
-    case COMP_LT:
-      if (stat_buf->st_ctime < (time_t) pred_ptr->args.info.l_val)
-	return (true);
-      break;
-    case COMP_EQ:
-      if ((stat_buf->st_ctime >= (time_t) pred_ptr->args.info.l_val)
-	  && (stat_buf->st_ctime < (time_t) pred_ptr->args.info.l_val + 60))
-	return (true);
-      break;
-    }
-  return (false);
+  (void) pathname;
+  return pred_timewindow(stat_buf->st_ctime, pred_ptr, 60);
 }
 
 boolean
 pred_cnewer (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
+  (void) &pathname;
+  
   if (stat_buf->st_ctime > pred_ptr->args.time)
     return (true);
   return (false);
@@ -383,29 +376,39 @@ pred_comma (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 boolean
 pred_ctime (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
-  switch (pred_ptr->args.info.kind)
+  (void) &pathname;
+  return pred_timewindow(stat_buf->st_ctime, pred_ptr, DAYSECS);
+}
+
+boolean
+pred_delete (pathname, stat_buf, pred_ptr)
+     char *pathname;
+     struct stat *stat_buf;
+     struct predicate *pred_ptr;
+{
+  if (strcmp (rel_pathname, "."))
     {
-    case COMP_GT:
-      if (stat_buf->st_ctime > (time_t) pred_ptr->args.info.l_val)
-	return (true);
-      break;
-    case COMP_LT:
-      if (stat_buf->st_ctime < (time_t) pred_ptr->args.info.l_val)
-	return (true);
-      break;
-    case COMP_EQ:
-      if ((stat_buf->st_ctime >= (time_t) pred_ptr->args.info.l_val)
-	  && (stat_buf->st_ctime < (time_t) pred_ptr->args.info.l_val
-	      + DAYSECS))
-	return (true);
-      break;
+      if (0 != remove (rel_pathname))
+	{
+	  error (0, errno, "cannot delete %s", pathname);
+	  return false;
+	}
+      else
+	{
+	  return true;
+	}
     }
-  return (false);
+  
+  /* nothing to do. */
+  return true;
 }
 
 boolean
 pred_empty (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
+  (void) &pathname;
+  (void) &pred_ptr;
+  
   if (S_ISDIR (stat_buf->st_mode))
     {
       DIR *d;
@@ -447,6 +450,9 @@ pred_empty (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 boolean
 pred_exec (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
+  (void) &pathname;
+  (void) &stat_buf;
+  
   int i;
   int path_pos;
   struct exec_val *execp;	/* Pointer for efficiency. */
@@ -485,6 +491,11 @@ pred_exec (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 boolean
 pred_false (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
+  (void) &pathname;
+  (void) &stat_buf;
+  (void) &pred_ptr;
+
+  
   return (false);
 }
 
@@ -499,6 +510,9 @@ pred_fls (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 boolean
 pred_fprint (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
+  (void) &pathname;
+  (void) &stat_buf;
+  
   fputs (pathname, pred_ptr->args.stream);
   putc ('\n', pred_ptr->args.stream);
   return (true);
@@ -507,6 +521,9 @@ pred_fprint (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 boolean
 pred_fprint0 (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
+  (void) &pathname;
+  (void) &stat_buf;
+  
   fputs (pathname, pred_ptr->args.stream);
   putc (0, pred_ptr->args.stream);
   return (true);
@@ -570,6 +587,11 @@ pred_fprintf (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 	case 'd':		/* depth in search tree */
 	  fprintf (fp, segment->text, curdepth);
 	  break;
+	case 'D':		/* Device on which file exists (stat.st_dev) */
+	  fprintf (fp, segment->text, 
+		   human_readable ((uintmax_t) stat_buf->st_dev, hbuf,
+				   human_ceiling, 1, 1));
+	  break;
 	case 'f':		/* basename of path */
 	  fprintf (fp, segment->text, base_name (pathname));
 	  break;
@@ -730,6 +752,53 @@ pred_fprintf (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 		   human_readable ((uintmax_t) stat_buf->st_uid, hbuf,
 				   human_ceiling, 1, 1));
 	  break;
+
+	/* type of filesystem entry like `ls -l`: (d,-,l,s,p,b,c,n) n=nonexistent(symlink) */
+	case 'Y':		/* in case of symlink */
+	  {
+#ifdef S_ISLNK
+	  if (S_ISLNK (stat_buf->st_mode))
+	    {
+	      struct stat sbuf;
+	      int (*ystat) ();
+	      ystat = xstat == lstat ? stat : lstat;
+	      if ((*ystat) (rel_pathname, &sbuf) != 0)
+	      {
+		if ( errno == ENOENT ) {
+		  fprintf (fp, segment->text, "N");
+		  break;
+		};
+		if ( errno == ELOOP ) {
+		  fprintf (fp, segment->text, "L");
+		  break;
+		};
+		error (0, errno, "", pathname);
+		/* exit_status = 1;
+		return (false); */
+	      }
+	      stat_buf->st_mode = sbuf.st_mode;
+	    }
+#endif /* S_ISLNK */
+	  }
+	  /* FALLTHROUGH */
+	case 'y':
+	  {
+	    mode_t m = stat_buf->st_mode & S_IFMT;
+
+	    fprintf (fp, segment->text,
+		( m == S_IFSOCK ? "s" :
+		  m == S_IFLNK  ? "l" :
+		  m == S_IFREG  ? "f" :
+		  m == S_IFBLK  ? "b" :
+		  m == S_IFDIR  ? "d" :
+		  m == S_IFCHR  ? "c" :
+#ifdef S_IFDOOR
+		  m == S_IFDOOR ? "D" :
+#endif
+		  m == S_IFIFO  ? "p" : "U" ) );
+
+	  }
+	  break;
 	}
     }
   return (true);
@@ -747,6 +816,8 @@ pred_fstype (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 boolean
 pred_gid (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
+  (void) pathname;
+  
   switch (pred_ptr->args.info.kind)
     {
     case COMP_GT:
@@ -768,6 +839,8 @@ pred_gid (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 boolean
 pred_group (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
+  (void) pathname;
+  
   if (pred_ptr->args.gid == stat_buf->st_gid)
     return (true);
   else
@@ -783,6 +856,8 @@ pred_ilname (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 boolean
 pred_iname (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
+  (void) stat_buf;
+
   const char *base;
 
   /* FNM_PERIOD is not used here because POSIX requires that it not be.
@@ -797,6 +872,8 @@ pred_iname (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 boolean
 pred_inum (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
+  (void) pathname;
+  
   switch (pred_ptr->args.info.kind)
     {
     case COMP_GT:
@@ -818,6 +895,8 @@ pred_inum (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 boolean
 pred_ipath (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
+  (void) stat_buf;
+  
   if (fnmatch (pred_ptr->args.str, pathname, FNM_CASEFOLD) == 0)
     return (true);
   return (false);
@@ -881,46 +960,15 @@ pred_ls (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 boolean
 pred_mmin (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
-  switch (pred_ptr->args.info.kind)
-    {
-    case COMP_GT:
-      if (stat_buf->st_mtime > (time_t) pred_ptr->args.info.l_val)
-	return (true);
-      break;
-    case COMP_LT:
-      if (stat_buf->st_mtime < (time_t) pred_ptr->args.info.l_val)
-	return (true);
-      break;
-    case COMP_EQ:
-      if ((stat_buf->st_mtime >= (time_t) pred_ptr->args.info.l_val)
-	  && (stat_buf->st_mtime < (time_t) pred_ptr->args.info.l_val + 60))
-	return (true);
-      break;
-    }
-  return (false);
+  (void) &pathname;
+  return pred_timewindow(stat_buf->st_mtime, pred_ptr, 60);
 }
 
 boolean
 pred_mtime (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
-  switch (pred_ptr->args.info.kind)
-    {
-    case COMP_GT:
-      if (stat_buf->st_mtime > (time_t) pred_ptr->args.info.l_val)
-	return (true);
-      break;
-    case COMP_LT:
-      if (stat_buf->st_mtime < (time_t) pred_ptr->args.info.l_val)
-	return (true);
-      break;
-    case COMP_EQ:
-      if ((stat_buf->st_mtime >= (time_t) pred_ptr->args.info.l_val)
-	  && (stat_buf->st_mtime < (time_t) pred_ptr->args.info.l_val
-	      + DAYSECS))
-	return (true);
-      break;
-    }
-  return (false);
+  (void) pathname;
+  return pred_timewindow(stat_buf->st_mtime, pred_ptr, DAYSECS);
 }
 
 boolean
@@ -1090,6 +1138,15 @@ pred_prune (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
   return (do_dir_first);	/* This is what SunOS find seems to do. */
 }
 
+boolean
+pred_quit (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
+{
+  (void) &pathname;
+  (void) &stat_buf;
+  (void) &pred_ptr;
+  exit(0);
+}
+
 boolean
 pred_regex (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
 {
@@ -1192,23 +1249,7 @@ pred_used (char *pathname, struct stat *stat_buf, struct predicate *pred_ptr)
   time_t delta;
 
   delta = stat_buf->st_atime - stat_buf->st_ctime; /* Use difftime? */
-  switch (pred_ptr->args.info.kind)
-    {
-    case COMP_GT:
-      if (delta > (time_t) pred_ptr->args.info.l_val)
-	return (true);
-      break;
-    case COMP_LT:
-      if (delta < (time_t) pred_ptr->args.info.l_val)
-	return (true);
-      break;
-    case COMP_EQ:
-      if ((delta >= (time_t) pred_ptr->args.info.l_val)
-	  && (delta < (time_t) pred_ptr->args.info.l_val + DAYSECS))
-	return (true);
-      break;
-    }
-  return (false);
+  return pred_timewindow(delta, pred_ptr, DAYSECS);
 }
 
 boolean
