@@ -57,6 +57,7 @@
 #include <time.h>
 #include <fnmatch.h>
 #include <getopt.h>
+#include <xstrtol.h>
 
 #define NDEBUG
 #include <assert.h>
@@ -105,6 +106,8 @@ extern int errno;
 
 #include "locatedb.h"
 #include <getline.h>
+#include "../gnulib/lib/xalloc.h"
+#include "../gnulib/lib/error.h"
 
 /* Note that this evaluates C many times.  */
 #ifdef _LIBC
@@ -129,10 +132,11 @@ static const char warn_name_units[] = N_("days");
 /* Check for existence of files before printing them out? */
 int check_existence = 0;
 
+/* What to separate the results with. */
+static int separator = '\n';
+
 char *next_element ();
-char *xmalloc ();
-char *xrealloc ();
-void error ();
+
 
 /* Read in a 16-bit int, high byte first (network byte order).  */
 
@@ -148,6 +152,33 @@ get_short (fp)
   return x;
 }
 
+const char * const metacharacters = "*?[]\\";
+
+/* Return nonzero if CH is a shell glob character.
+ */
+static int 
+is_metacharacter(char ch)
+{
+  if (NULL == strchr (metacharacters, ch))
+    return 0;
+  else
+    return 1;
+}
+
+/* Return nonzero if S contains any shell glob characters.
+ */
+static int 
+contains_metacharacter(const char *s)
+{
+  if (NULL == strpbrk(s, metacharacters))
+    return 0;
+  else
+    return 1;
+}
+
+  
+
+#if 0
 /* Return a pointer to the last character in a static copy of the last
    glob-free subpattern in NAME,
    with '\0' prepended for a fast backwards pre-match.  */
@@ -163,7 +194,7 @@ last_literal_end (name)
 
   /* Find the end of the subpattern.
      Skip trailing metacharacters and [] ranges. */
-  for (p = name + strlen (name) - 1; p >= name && strchr ("*?]", *p) != NULL;
+  for (p = name + strlen (name) - 1; p >= name && is_metacharacter(*p);
        p--)
     {
       if (*p == ']')
@@ -183,13 +214,13 @@ last_literal_end (name)
 
   /* If the pattern has only metacharacters, make every path match the
      subpattern, so it gets checked the slow way.  */
-  if (p == name && strchr ("?*[]", *p) != NULL)
+  if (p == name && is_metacharacter(*p))
     *subp++ = '/';
   else
     {
       char *endmark;
       /* Find the start of the metacharacter-free subpattern.  */
-      for (endmark = p; p >= name && strchr ("]*?", *p) == NULL; p--)
+      for (endmark = p; p >= name && !is_metacharacter(*p); p--)
 	;
       /* Copy the subpattern into globfree.  */
       for (++p; p <= endmark; )
@@ -199,6 +230,8 @@ last_literal_end (name)
 
   return subp;
 }
+#endif 
+
 
 /* locate_read_str()
  *
@@ -247,55 +280,231 @@ locate_read_str(char **buf, size_t *siz, FILE *fp, int delimiter, int offs)
 }
 
 
+static void
+lc_strcpy(char *dest, const char *src)
+{
+  while (*src)
+    {
+      *dest++ = TOLOWER(*src);
+      ++src;
+    }
+}
+
+enum visit_result
+  {
+    VISIT_CONTINUE = 1,  /* please call the next visitor */
+    VISIT_ACCEPTED = 2,  /* accepted, call no futher callbacks for this file */
+    VISIT_REJECTED = 4,  /* rejected, process next file. */
+    VISIT_ABORT    = 8   /* rejected, process no more files. */
+  };
+
+
+
+struct casefolder
+{
+  const char *pattern;
+  char *buffer;
+  size_t buffersize;
+};
+
+
+
+typedef int (*visitfunc)(const char *filename, void *context);
+
+struct visitor
+{
+  visitfunc      inspector;
+  void *         context;
+  struct visitor *next;
+};
+
+
+static struct visitor *inspectors = NULL;
+static struct visitor *lastinspector = NULL;
+
+static int
+process_filename(const char *filename)
+{
+  int result = VISIT_CONTINUE;
+  const struct visitor *p = inspectors;
+  
+  while ( (VISIT_CONTINUE == result) && (NULL != p) )
+    {
+      result = (p->inspector)(filename, p->context);
+      p = p->next;
+    }
+
+  if (VISIT_CONTINUE == result)
+    return VISIT_ACCEPTED;
+  else
+    return result;
+}
+
+static void
+add_visitor(visitfunc fn, void *context)
+{
+  struct visitor *p = xmalloc(sizeof(struct visitor));
+  p->inspector = fn;
+  p->context   = context;
+  p->next = NULL;
+  
+  if (NULL == lastinspector)
+    {
+      lastinspector = inspectors = p;
+    }
+  else
+    {
+      lastinspector->next = p;
+    }
+}
+
+
+
+static int
+visit_justprint(const char *name, void *context)
+{
+  (void)&context;
+  fputs(name, stdout);
+  putchar(separator);
+  return VISIT_CONTINUE;
+}
+
+static int
+visit_exists(const char *name, void *context)
+{
+  struct stat st;
+  (void)&context;
+  if (stat(name, &st) != 0)
+    {
+      return VISIT_REJECTED;
+    }
+  else
+    {
+      return VISIT_CONTINUE;
+    }
+}
+
+static int
+visit_substring_match_nocasefold(const char *name, void *context)
+{
+  const char *pattern = context;
+
+  if (NULL != strstr(name, pattern))
+    return VISIT_CONTINUE;
+  else
+    return VISIT_REJECTED;
+}
+
+static int
+visit_substring_match_casefold(const char *name, void *context)
+{
+  struct casefolder * p = context;
+  size_t len = strlen(name);
+
+  if (len > p->buffersize)
+    {
+      p->buffer = xrealloc(p->buffer, len+1);
+      p->buffersize = len+1;
+    }
+  lc_strcpy(p->buffer, name);
+  
+  
+  if (NULL != strstr(p->buffer, p->pattern))
+    return VISIT_CONTINUE;
+  else
+    return VISIT_REJECTED;
+}
+
+
+static int
+visit_globmatch_nofold(const char *name, void *context)
+{
+  const char *glob = context;
+  if (fnmatch(glob, name, 0) != 0)
+    return VISIT_REJECTED;
+  else
+    return VISIT_CONTINUE;
+}
+
+
+static int
+visit_globmatch_casefold(const char *name, void *context)
+{
+  const char *glob = context;
+  if (fnmatch(glob, name, FNM_CASEFOLD) != 0)
+    return VISIT_REJECTED;
+  else
+    return VISIT_CONTINUE;
+}
+
 /* Print the entries in DBFILE that match shell globbing pattern PATHPART.
    Return the number of entries printed.  */
 
-static int
-locate (pathpart, dbfile, ignore_case)
-     char *pathpart, *dbfile;
-     int ignore_case;
+static unsigned long
+new_locate (char *pathpart, char *dbfile, int ignore_case, int enable_print, int basename_only, int use_limit, uintmax_t limit)
 {
-  /* The pathname database.  */
-  FILE *fp;
-  /* An input byte.  */
-  int c;
-  /* Number of bytes read from an entry.  */
-  int nread;
+  
+  FILE *fp;			/* The pathname database.  */
+  int c;			/* An input byte.  */
+  int nread;		     /* number of bytes read from an entry. */
+  char *path;		       /* The current input database entry. */
+  size_t pathsize;		/* Amount allocated for it.  */
+  int count = 0; /* The length of the prefix shared with the previous database entry.  */
+  
+  boolean old_format = false; /* true if reading a bigram-encoded database.  */
+  
 
-  /* true if PATHPART contains globbing metacharacters.  */
-  boolean globflag;
-  /* The end of the last glob-free subpattern in PATHPART.  */
-  char *patend;
 
-  /* The current input database entry.  */
-  char *path;
-  /* Amount allocated for it.  */
-  size_t pathsize;
-
-  /* The length of the last path.  */
-  size_t last_pathlen;
-
-  /* The length of the prefix shared with the previous database entry.  */
-  int count = 0;
-  /* Where in `path' to stop the backward search for the last character
-     in the subpattern.  Set according to `count'.  */
-  char *cutoff;
-
-  /* true if we found a fast match (of patend) on the previous path.  */
-  boolean prev_fast_match = false;
-  /* The return value.  */
-  int printed = 0;
-
-  /* true if reading a bigram-encoded database.  */
-  boolean old_format = false;
-  /* For the old database format,
+  /* for the old database format,
      the first and second characters of the most common bigrams.  */
   char bigram1[128], bigram2[128];
 
+  /* number of items accepted (i.e. printed) */
+  unsigned long int items_accepted = 0uL;
+
   /* To check the age of the database.  */
   struct stat st;
   time_t now;
 
+  /* Set up the inspection regime */
+  inspectors = NULL;
+  lastinspector = NULL;
+
+  if (check_existence)
+    add_visitor(visit_exists, NULL);
+
+  if (contains_metacharacter(pathpart))
+    {
+      if (ignore_case)
+	add_visitor(visit_globmatch_casefold, pathpart);
+      else
+	add_visitor(visit_globmatch_nofold, pathpart);
+    }
+  else
+    {
+      /* No glob characters reuired.  Hence we match on 
+       * _any part_ of the filename, not just the 
+       * basename.  This seems odd to me, but it was 
+       * the behaviour of the code before I touched it...
+       * James Youngman <jay@gnu.org> 
+       */
+      if (ignore_case)
+	{
+	  struct casefolder * cf = xmalloc(sizeof(*cf));
+	  cf->pattern = pathpart;
+	  cf->buffer = NULL;
+	  cf->buffersize = 0;
+	  add_visitor(visit_substring_match_casefold, cf);
+	}
+      else
+	{
+	  add_visitor(visit_substring_match_nocasefold, pathpart);
+	}
+    }
+
+  if (enable_print)
+    add_visitor(visit_justprint, NULL);
+  
   if (stat (dbfile, &st) || (fp = fopen (dbfile, "r")) == NULL)
     {
       error (0, errno, "%s", dbfile);
@@ -328,22 +537,18 @@ locate (pathpart, dbfile, ignore_case)
       old_format = true;
     }
 
-  /* If we ignore case,
-     convert it to lower first so we don't have to do it every time */
-  if (ignore_case){
-    for (patend=pathpart;*patend;++patend){
-     *patend=TOLOWER(*patend);
+  /* If we ignore case, convert it to lower first so we don't have to
+   * do it every time
+   */
+  if (ignore_case)
+    {
+      lc_strcpy(pathpart, pathpart);
     }
-  }
   
-  
-  globflag = strchr (pathpart, '*') || strchr (pathpart, '?')
-    || strchr (pathpart, '[');
-
-  patend = last_literal_end (pathpart);
+  items_accepted = 0;
 
   c = getc (fp);
-  while (c != EOF)
+  while ( (c != EOF) && (!use_limit || (limit > 0)) )
     {
       register char *s;		/* Scan the path we read in.  */
 
@@ -377,8 +582,6 @@ locate (pathpart, dbfile, ignore_case)
 	  else
 	    count += c;
 
-	  assert(count >= 0 && count <= last_pathlen);
-	  
 	  if (count > strlen(path))
 	    {
 	      /* This should not happen generally , but since we're
@@ -399,55 +602,13 @@ locate (pathpart, dbfile, ignore_case)
 	  assert (s[2] == '\0'); /* Added by locate_read_str.  */
 	}
 
-      /* If the previous path matched, scan the whole path for the last
-	 char in the subpattern.  If not, the shared prefix doesn't match
-	 the pattern, so don't scan it for the last char.  */
-      cutoff = prev_fast_match ? path : path + count;
-
-      /* Search backward starting at the end of the path we just read in,
-	 for the character at the end of the last glob-free subpattern
-	 in PATHPART.  */
-      for(prev_fast_match=false; s>=cutoff; s--)
-        {
-          char *s2;   /* Scan the path we read in. */
-          register char *p2;  /* Scan `patend'.  */
-
-          /* Fast first char check. */
-          if(ignore_case)
-            {
-              if(TOLOWER(*s)!=*patend)
-                continue;
-            }
-          else if(*s!=*patend)
-            continue;
-
-          if(ignore_case)
-            for(s2=s-1, p2=patend-1; *p2!='\0' && TOLOWER(*s2)==*p2; s2--, p2--);
-          else
-            for(s2=s-1, p2=patend-1; *p2!='\0' && *s2==*p2; s2--, p2--);
-
-          if(*p2!='\0')
-            continue;
-          /*  Success on the fast match.  Compare the whole pattern
-              if it contains globbing characters.  */
-          prev_fast_match = true;
-          if(globflag)
-            {
-              if(ignore_case)
-                {
-                  if(fnmatch(pathpart,basename(path),FNM_CASEFOLD)!=0)
-                    break;
-                }
-              else
-                if(fnmatch(pathpart,basename(path),0)!=0)
-                  break;
-            }
-          if(check_existence && stat(path,&st)!=0)
-            break;
-          puts(path);
-          ++printed;
-          break;
-        }
+      if (VISIT_ACCEPTED == process_filename(basename_only ? basename(path) : path))
+	{
+	  if ((++items_accepted >= limit) && use_limit)
+	    {
+	      break;
+	    }
+	}
     }
   
   if (ferror (fp))
@@ -461,9 +622,12 @@ locate (pathpart, dbfile, ignore_case)
       return 0;
     }
 
-  return printed;
+  return items_accepted;
 }
 
+
+
+
 extern char *version_string;
 
 /* The name this program was run with. */
@@ -475,11 +639,33 @@ usage (stream)
 {
   fprintf (stream, _("\
 Usage: %s [-d path | --database=path] [-e | --existing]\n\
-      [-i | --ignore-case] [--version] [--help] pattern...\n"),
+      [-i | --ignore-case] [--wholepath] [--basename] [--limit=N | -l N]\n\
+      [--version] [--help] pattern...\n"),
 	   program_name);
   fputs (_("\nReport bugs to <bug-findutils@gnu.org>.\n"), stream);
 }
 
+static void
+sanity_check_dbpath(const char *path)
+{
+  size_t len;
+  
+  if (':' == path[0])
+    {
+      error(0, 0, _("warning: locate database path `%s' contains a leading colon, which is not a valid database name"));
+    }
+
+  len = strlen(path);
+  if (len > 0u)
+    {
+      if (':' == path[len-1u])
+	{
+	  error(0, 0, _("warning: locate database path `%s' contains a trailing colon, which is not a valid database name"));
+	}
+    }
+}
+
+
 static struct option const longopts[] =
 {
   {"database", required_argument, NULL, 'd'},
@@ -487,6 +673,13 @@ static struct option const longopts[] =
   {"ignore-case", no_argument, NULL, 'i'},
   {"help", no_argument, NULL, 'h'},
   {"version", no_argument, NULL, 'v'},
+  {"null", no_argument, NULL, '0'},
+  {"count", no_argument, NULL, 'c'},
+  {"wholename", no_argument, NULL, 'w'},
+  {"basename", no_argument, NULL, 'b'},
+  {"stdio", no_argument, NULL, 's'},
+  {"mmap",  no_argument, NULL, 'm'},
+  {"limit",  required_argument, NULL, 'l'},
   {NULL, no_argument, NULL, 0}
 };
 
@@ -496,10 +689,15 @@ main (argc, argv)
      char **argv;
 {
   char *dbpath;
-  int fnmatch_flags = 0;
-  int found = 0, optc;
+  unsigned long int found = 0uL;
+  int optc;
   int ignore_case = 0;
-
+  int print = 1;
+  int just_count = 0;
+  int basename_only = 0;
+  uintmax_t limit = 0;
+  int use_limit = 0;
+  
   program_name = argv[0];
 
 #ifdef HAVE_SETLOCALE
@@ -514,9 +712,22 @@ main (argc, argv)
 
   check_existence = 0;
 
-  while ((optc = getopt_long (argc, argv, "d:ei", longopts, (int *) 0)) != -1)
+  while ((optc = getopt_long (argc, argv, "cd:eil:sm0", longopts, (int *) 0)) != -1)
     switch (optc)
       {
+      case '0':
+	separator = 0;
+	break;
+
+      case 'b':
+	basename_only = 1;
+	break;
+
+      case 'c':
+	just_count = 1;
+	print = 0;
+	break;
+
       case 'd':
 	dbpath = optarg;
 	break;
@@ -527,7 +738,6 @@ main (argc, argv)
 
       case 'i':
 	ignore_case = 1;
-	fnmatch_flags |= FNM_CASEFOLD;
 	break;
 	
       case 'h':
@@ -538,6 +748,29 @@ main (argc, argv)
 	printf (_("GNU locate version %s\n"), version_string);
 	return 0;
 
+      case 'w':
+	basename_only = 0;
+	break;
+
+      case 'l':
+	{
+	  char *end = optarg;
+	  strtol_error err = xstrtoumax(optarg, &end, 10, &limit, NULL);
+	  if (LONGINT_OK != err)
+	    {
+	      STRTOL_FATAL_ERROR(optarg, _("argument to --limit"), err);
+	    }
+	  use_limit = 1;
+	}
+	break;
+
+      case 's':			/* use stdio */
+      case 'm':			/* use mmap  */
+	/* These options are implemented simply for
+	 * compatibility with FreeBSD
+	 */ 
+	break;
+
       default:
 	usage (stderr);
 	return 1;
@@ -548,17 +781,25 @@ main (argc, argv)
       usage (stderr);
       return 1;
     }
-  
+
+  sanity_check_dbpath(dbpath);
 
   for (; optind < argc; optind++)
     {
       char *e;
       next_element (dbpath);	/* Initialize.  */
       while ((e = next_element ((char *) NULL)) != NULL)
-	found |= locate (argv[optind], e, ignore_case);
+	{
+	  found += new_locate (argv[optind], e, ignore_case, print, basename_only, use_limit, limit);
+	}
     }
 
-  if (found)
+  if (just_count)
+    {
+      printf("%ld\n", found);
+    }
+  
+  if (found || (use_limit && (limit==0)) )
     return 0;
   else
     return 1;
